import clr
clr.AddReferenceToFileAndPath(r"C:\Program Files\ANSYS Inc\v192\Addins\ACT\bin\Win64\MathNet.Numerics.dll")     # Using the Math.NET dll. Can not use NumPy, as this is not supported by IronPython.
import MathNet.Numerics.LinearAlgebra as la    # math library
import os
import subprocess
import datetime         # For writing current date in generated FTL file
import ctypes
import cPickle as pickle       # For saving object to file
import System


def createUnitVector(pos, nnodes, cmodes=0):
    '''
    Creates a vector with unit length at the specified position pos. 

    :param pos: Int of position to give unit length. 1=UX, 2=UY, 3=UZ, 4=ROTX, 5=ROTY, 6=ROTZ
    :param nnodes: Int of number of nodes to create unit length for.
    :param cmodes: Number of component nodes to add unit length
    return  matrix of size ((nnodes*6)+cmodes x 1)
    '''
    if (pos <= 0):
        raise Exception("pos argument must be int > 0")
        
    u = la.Matrix[System.Double].Build.Dense(nnodes*6 + cmodes, 1)     # initializing with zeroes
    for i in range(pos-1, nnodes*6, 6):     # Adding 1 at position pos
        u[i, 0] = 1
    for i in range(nnodes*6, nnodes*6 + cmodes):        # Filling component mode entries with 0s
        u[i, 0] = 0
    return u



    
def onCreateModRed(analysis):
    """
    Adding the ModRed model object in the model tree
    
    :param analysis: The currently active analysis
    """
    if str(analysis.AnalysisType) == "Modal":
        # Use the analysis to create the APDL Based Result Evaluation result.
        analysis.CreateLoadObject("modred", "ModRed")

    else:
        # Display an error message in the Mechanical message log.
        ExtAPI.Application.LogError("Can only select ModRed for Modal analysis.")


def saveObject(obj, fname):
    """
    Saves the current object to _fedemDir_
    """
    testResourcesPath = ExtAPI.ExtensionManager.CurrentExtension.InstallDir + "\\test-resources"
    with open(testResourcesPath + "\\" + fname + ".p", "wb") as input: 
        pickle.dump(obj, input, pickle.HIGHEST_PROTOCOL)


def onExportData(analysis):       
    '''
    :param analysis: Ansys.ACT.Automation.Mechanical.Analysis object.
    - Calculates gravity vectors from reduced mass matrix
    - Creates FTL file for Fedem
    - Creates FMX files from fmxWriter DLL
    - All files are saved in solverData._fedemDir
    '''

    ExtAPI.Log.WriteMessage("Exporting data for Fedem...")
    load = analysis.GetLoadObjects("ModRed")[0]
    solverData = SolverData(analysis, load)           # Creating a SolverData object for the current analysis.
    
    # Collecting exported matrices in Math.NET format
    M_red = readMMFMatrix(analysis.WorkingDir + "M_red.mmf", "sub")
    K_red = readMMFMatrix(analysis.WorkingDir + "K_red.mmf", "sub")

    G = calculateGravityVectorReduced(M_red, len(solverData._eNodes), int(solverData._cModes))  # Using reduced mass matrix
    # Generating FTL file for Fedem
    ftlFile = open(os.path.join(solverData._fedemDir, solverData._filename + ".ftl"), "w")        # Creating a new ftl file
    generateFTL(solverData.getNodes(), solverData.getElements(), solverData._eNodes, ftlFile)       # Generating FTL file to be placed in the solver directory.
    ExtAPI.Log.WriteMessage("FTL file " + str(solverData._filename) + ".ftl written to " + solverData._fedemDir)

    # Generating FMX file for Fedem
    generateFMX(mathNET2list(G), mathNET2list(M_red), mathNET2list(K_red), solverData)
    
def calculateGravityVectorReduced(M_red, nENodes, cmodes):
    """
    Calculates gravity vectors for Fedem from the reduced mass matrix.
    :param M_red: The reduced mass matrix.
    :param nENodes: Number of selected external nodes
    """
    # Creating reduced unit vectors
    u_x = createUnitVector(1, nENodes, cmodes)
    u_y = createUnitVector(2, nENodes, cmodes)
    u_z = createUnitVector(3, nENodes, cmodes)
    u = u_x.Append(u_y).Append(u_z)

    G_red = M_red.Multiply(u)    
    return G_red




    
def calculateGravityVector(M_full, H, nENodes, nINodes, dofs):
    """
    Calculates gravity vectors for Fedem from the full mass matrix and H matrix.
    Follows equations from page 118 in "Virtual Testing of Mechanical Systems"
    :param M_full: The full, partitioned mass matrix.
    :param H: The H matrix
    H = 
    [[I 0],
    [B, phi]]
    :param nENodes: Number of selected external nodes
    :param nINodes: Number of selected internal nodes
    :param dofs: Number of degrees of freedom for the element
    """
    # Extracting the partitions of the M matrix
    partitions = extractPartitions(M_full, nENodes, dofs)
    M_ee = partitions[0]
    M_ei = partitions[1]
    M_ie = partitions[2]
    M_ii = partitions[3]

    # Creating unit vectors:
    u_x_e = createUnitVector(1, nENodes)
    u_x_i = createUnitVector(1, nINodes)
    u_y_e = createUnitVector(2, nENodes)
    u_y_i = createUnitVector(2, nINodes)
    u_z_e = createUnitVector(3, nENodes)
    u_z_i = createUnitVector(3, nINodes)

   # Gravitational forces: 
    G_x_i = M_ii.Multiply(u_x_i).Add(M_ie.Multiply(u_x_e))      # M_ii*u_x_i + M_ie * u_x_e
    G_x_e = M_ei.Multiply(u_x_i).Add(M_ee.Multiply(u_x_e))
    G_y_i = M_ii.Multiply(u_y_i).Add(M_ie.Multiply(u_y_e))
    G_y_e = M_ei.Multiply(u_y_i).Add(M_ee.Multiply(u_y_e))
    G_z_i = M_ii.Multiply(u_z_i).Add(M_ie.Multiply(u_z_e))
    G_z_e = M_ei.Multiply(u_z_i).Add(M_ee.Multiply(u_z_e))
    
    # Collecting in one matrix: 
    G_x = G_x_e.Stack(G_x_i)
    G_y = G_y_e.Stack(G_y_i)
    G_z = G_z_e.Stack(G_z_i)
    
    # Reducing to unit gravitational acceleration on the full matrix: 
    G_x = H.Transpose().Multiply(G_x)
    G_y = H.Transpose().Multiply(G_y)
    G_z = H.Transpose().Multiply(G_z)
    
    # Collecting gravity vectors in one matrix
    G = G_x.Append(G_y).Append(G_z)
    return G



def extractPartitions(mat, numENodes, dofs):
    """
    Extracts the partitions 
    M =
    [[M_ee, M_ei],
    [M_ie, M_ii]]
    from the matrix mat
    M_ee of size (numENodes*dofs, numENodes*dofs)

    :param mat: Symmetric Math.NET matrix already partitioned into M_ee, M_ei, ... 
    :param eNodes: Number of selected external nodes
    :param dofs: Number of degrees of freedom per node
    """
    nRows = mat.RowCount
    nCols = mat.ColumnCount
    
    M_ee = mat.SubMatrix(0, numENodes*dofs, 0, numENodes*dofs)      # Top left of M
    M_ei = mat.SubMatrix(0, numENodes*dofs, numENodes*dofs, nCols - numENodes * dofs)     # Top right
    M_ie = mat.SubMatrix(numENodes*dofs, nRows - numENodes * dofs, 0, numENodes*dofs)     # Bottom left
    M_ii = mat.SubMatrix(numENodes*dofs, nRows - numENodes * dofs, numENodes*dofs, nCols - numENodes*dofs)        # Bottom right
    return [ M_ee, M_ei, M_ie, M_ii ]

def mathNET2list(matrix):
    """
    Returns a Python list of the items in the matrix

    :param matrix: Math.NET matrix
    return One-dimensional list of items in matrix saved column-first.
    """
    list = [0] * (matrix.ColumnCount * matrix.RowCount)
    index = 0
    for col in range(0, matrix.ColumnCount):
        for row in range(0, matrix.RowCount):
            list[index] = matrix[row, col]
            index += 1
    return list
            


def onSolve(load, stream):
    """
    Adds the required APDL commands to the solver input (ds.dat) file. Activates when the user hits "Solve"
     
    :param load: the load associated to the callback. Interface IUserLoad
    :param stream: a System.IO.StringWriter object, to which solver commands should be appended (represents the ds.dat file)
    """

    # Collecting user input:
    nodeSelectionMethod = load.Properties["Generation"].Properties["Geometry"].Value
    method = load.Properties["Method"].Properties["ReductionMethod"].Value
    filename = load.Properties["Method"].Properties["ReductionMethod"].Properties["Name"].Value  # string
    nmode = load.Properties["Method"].Properties["ReductionMethod"].Properties["Nmode"].Value     # string
    isUseLumped = load.Properties["Method"].Properties["ReductionMethod"].Properties["LumpedMatrix"].Value     # string
    eNodes = []      
    if (nodeSelectionMethod=="Geometry"):
        eNodes = getENodes(load)      # List[int]
    if (nodeSelectionMethod=="RBE2 Import"):
        eNodes = getRBE2Nodes(load)
    isExport = load.Properties["Export"].Properties["IsExport"].Value
    if (isExport == "Yes"):
        isExport = True
    else: 
        isExport = False

    stiffnessName = "Kmat"
    massName = "Mmat"
    matrixFormat = "MMF"        # Set this to ".matrix" for lower precision, but more readable matrices
    
    #
    # Writing solver commands to the ds.dat file
    #
    stream.WriteLine("finish")
    stream.WriteLine("/filname, " + filename)     # Name of the super element file to be generated
    stream.WriteLine("save")
    stream.WriteLine("/solu")
    stream.WriteLine("antype, substr")          # Defining a substructure analysis type (includes both Guyan and CMS)
    if (method == "CMS"):  # If CMSOPT needs to be set
        cmsmeth = "fix"
        nmode = str(nmode)
        freqb = ""
        freqe = ""
        fbddef = ""
        fbdval = ""
        iokey = "tcms"
        if (iokey=="tcms"):
            stream.WriteLine("outpr, nsol, all")        # Print command that must be defined if iokey==tcms
        stream.WriteLine("cmsopt, " + cmsmeth + "," + nmode + "," + freqb + "," + freqe + "," + fbddef + "," + fbdval + "," + iokey)
    sename = filename           # name of superelement matrix file
    sematr = str(2)          # generate stiffness and mass matrices
    sepr = str(0)            # Do not print superelement matrices or load vectors
    sesst = str(0)       # Do not save space for stress stiffening
    stream.WriteLine("seopt, " + sename + ", " + sematr + ", " + sepr + ", " + sesst + ", ,")
    for nodeID in eNodes: 
        stream.WriteLine("m, " + str(nodeID) + ", all")
    if(isUseLumped == "Yes"):
        stream.WriteLine("lumpm, on")           # Using a lumped mass matrix
    stream.WriteLine("alls")
    stream.WriteLine("solve")
    stream.WriteLine("save")
    stream.WriteLine("finish")
    stream.WriteLine("! Master node selection finished")
    
    if (isExport):              # Exporting matrices if selected
        stream.WriteLine("! Exporting matrices...")
        # Creating H matrix. H = [CST NOR]
        stream.WriteLine("*dmat, cst, D, import, tcms, " + filename + ".tcms, cst")          # Constraint mode data
        stream.WriteLine("*dmat, nor, D, import, tcms, " + filename + ".tcms, nor")          # Fixed-interface normal mode data
        stream.WriteLine("save")
        if (matrixFormat =="MMF"):
            stream.WriteLine("*export, cst, MMF, CST.mmf")              # Exporting constraint modes
            stream.WriteLine("*export, nor, MMF, NOR.mmf")              # Exporting fixed-interface normal mode data
            
            stream.WriteLine("*smat, nod2solv, D, import, full, " + filename + ".full, nod2solv")       #  Importing the mapping vector internal -> solver ordering
            stream.WriteLine("*vec, mapback, I, import, full, " + filename + ".full, back")     # Importing the BACK nodal mapping vector for  external -> internal ordering.

            stream.WriteLine("/aux2")           # Manipulating binary files
            stream.WriteLine("FILEAUX2, " + filename + ", sub")          # Specifying to dump reduced matrices file
            stream.WriteLine("HBMAT, M_red, hbmat, , ASCII, MASS, NO")           # Dumping reduced mass matrix to Harwell-Boeing format for 16-decimal precision
            stream.WriteLine("HBMAT, K_red, hbmat, , ASCII, STIFF, NO")     # Dumping reduced stiffness matrix      
            
            stream.WriteLine("FILEAUX2, " + filename + ", full")       # Specifying to dump full mass matrix
            stream.WriteLine("HBMAT, M_full, hbmat, , ASCII, MASS, NO, YES")       # Dumping full mass matrix in Harwell-Boeing format
            stream.WriteLine("finish")
            stream.WriteLine("*smat, M_red, D, import, HBMAT, M_red.hbmat, ASCII")    # Creating sparse matrix from Harwell-Boeing format. Preserving 16-decimal precision
            stream.WriteLine("*smat, M_full, D, import, HBMAT, M_full.hbmat, ASCII")
            stream.WriteLine("*smat, K_red, D, import, HBMAT, K_red.hbmat, ASCII")
            
            stream.WriteLine("*export, M_red, MMF, M_red.mmf")            # Exporting to Matrix Market Format. Preserves 16-decimal precision
            stream.WriteLine("*export, M_full, MMF, M_full.mmf")
            stream.WriteLine("*export, K_red, MMF, K_red.mmf")            
                       
        if (matrixFormat == ".matrix"):
            stream.WriteLine("*PRINT, Kmat, " + stiffnessName + ".matrix")         # Printing matrix directly in easy-to-read format
            stream.WriteLine("*PRINT, Mmat, " + massName + ".matrix")     

        stream.WriteLine("save")

    stream.WriteLine("/eof")        # Quits correctly
    



def generateFMX(G, M_red, K_red, solverData):
    """
    This will save FMX files of G, M_red and K_red at the solver directory.

    :param G: One-dimensional list of gravitational vectors saved column-first.
    :param M_red: One-dimensional list of reduced mass matrix saved column-first.
    :param K_red: One-dimensional list of reduced stiffness matrix saved column-first.
    :param solverData: SolverData object.

    """
    # Converting to double slashes and in bytes representation
    file = solverData._fedemDir + "\\" + solverData._filename
    file = file.replace("\\", "\\\\")
    file = bytes(file, 'utf-8')

    # Skriver stivhetsmatrise
    status = writeFMX(file, 1, K_red)

    # Skriver massematrise
    status = writeFMX(file, 2, M_red)

    # Skriver gravitasjonskrefter
    status = writeFMX(file, 3, G)

    ExtAPI.Log.WriteMessage("FMX files written to " + solverData._fedemDir)



def writeFMX(file, ityp, data):
    """
    Writes a square matrix as a binary FMX-file for FEDEM, using the fmxWriter DLL

    arg file : Full path to the fmx-file to be written
    arg ityp : 1=stiffness matrix, 2=mass matrix, 3=gravity force vectors
    arg data : Matrix content, column-wise storage
    return : Zero on success, otherwise negative
    """
    # Loading DLL for writing FMX files:
    extensionDir = ExtAPI.ExtensionManager.CurrentExtension.InstallDir
    dll = ctypes.cdll.LoadLibrary(extensionDir + "\\fmxWriter\\fmxWriter.dll")
    cfil = file
    ctyp = ctypes.c_int(ityp)
    cdat = (ctypes.c_double*len(data))()
    cdat[:] = data
    clen = ctypes.c_int(len(data))
    return dll.WRITEFMX(cfil, ctypes.byref(ctyp), cdat, ctypes.byref(clen), len(file))


def getENodes(load):
    """   
    Returning list with node numbers of selected external nodes
    
    :param load: The current load object
    :return List[int] Sorted list with integers representing node numbers of current selected nodes. 
    """
    selectedIDs = load.Properties["Generation"].Properties["Geometry"].Properties["Geometry"].Value.Ids        
    return sorted(selectedIDs)      # List[int]



def generateFTL(nodes, elements, eNodes, file):
    '''
    Generates the FTL file for the current model to be used by Fedem. 
    The FTL file is placed in the solverData._fedemDir_

    :param nodes: list[Node] of all nodes in the model. Sorted ascending by id
    :param elements: list[Element] of all elements in the model. Sorted ascending by id
    :param eNodes: list[int] of ids of external nodes
    :param file: FTL file object
    '''
    #
    # Writing file header
    #
    date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    file.write("FTLVERSION{4 ASCII}\n")
    string ="#\n# FTL file generated by ModRed at " + str(date) + "\n#\n\n"
    file.write(string)
    string = "#\n# Nodal coordinates\n#"
    file.write(string + "\n")

    #
    # Writing NODE part
    #
    for node in nodes:
        nodeString = getNodeString(node, eNodes)
        file.write(nodeString + "\n")

    #
    # Writing element definitions part
    #
    thicknessId = 1
    materialId = 1
    file.write("#\n# Element definitions\n#\n")
    for element in elements:
        elString = getElementString(element, thicknessId, materialId)
        file.write(elString + "\n")
    #
    # Writing coordinate system part
    #
    file.write("#\n# Local coordinate systems\n#\n")
    file.write("PCOORDSYS{2 0 0 0 0 0 1 1 0 0}" + "\n")         

    #
    # Writing material properties part
    #
    file.write("#\n# Material properties\n#\n")
    file.write("PMAT{1 2.07e+11 8.02e+10 0.29 7820}" + "\n")       # Dummy material

    #
    # Writing shell thicknesses part
    #
    file.write("#\n# Shell thicknesses\n#\n")
    file.write("PTHICK{1 0.02}" + "\n")

    #
    # End of file
    #
    file.write("#\n# End of file")

    file.close()            # Closing the created FTL file
    

def getNodeString(node, eNodes):
    """
    Returns node string to write in FTL file. 

    :param id: id of node to write to string
    :param nodes: sorted list[Node] of all nodes in the model. Sorted by node id
    :param eNodes: list[int] of ids of selected external nodes
    :return string of NODE part to be written in FTL file
    """
    string = "NODE{"
    string += str(node._id)
    string += " "
    if(node._id in eNodes):
        string += "1"
    else: 
        string += "0"
    string += " "
    string += str(node._x) + " " + str(node._y) + " " + str(node._z)
    string += "}"
    return string


def getElementString(element, thicknessId, materialId):
    """
    Returns string of ELEMENT part to be written in FTL file

    :param element: Element object of the element to write
    :param thicknessID: id of thickness PTHICK
    :param materialID: id of material PMAT
    """
    string = "QUAD4{"
    string += str(element._id) + " "
    string += str(element._nodes[0]._id) + " "
    string += str(element._nodes[1]._id) + " "
    string += str(element._nodes[2]._id) + " "
    string += str(element._nodes[3]._id) + " "
    string += "{PTHICK "
    string += str(thicknessId) + "}"
    string += " "
    string += "{PMAT "
    string += str(materialId) + "}"
    string += "}"
    
    return string
    



def isValidNoModes(entity, property):
    '''
    Returns true if valid input to component modes property. False if not. 
    
    :param entity: IUserObject. The entity containing the property.
    :param property: ISimProperty. The property to check.
    
    :return True if property is positive int. False if not.
    '''
    try: 
        #ExtAPI.Log.WriteMessage(str(int(property.Value) >= 0))
        return int(property.Value) >= 0
    except ValueError:
        #ExtAPI.Log.WriteMessage(str(False))
        return False
        
def readMappingFileFedem(path):
    """
    Reads a Fedem mapping file that maps degree of freedom to equation. 
    The MEQN file is structured 
    column 0: dof       column 1: equation
    
    1   25
    2   26
    3   27  ...

    This means dof 1 equals equation 25.

    :param path: File path of MEQN.res file from Fedem
    :return Dictionary on the form
    dict = {
    "dof": equation
    }
    """
    file = open(path, "r")
    line = file.readline()          # Reading header line
    dict = {}
    lines = file.readlines()
    for line in lines:
        line = line.split()
        dof = int(line[0])
        eq = int(line[1])
        dict[dof] = eq
    return dict

def readMappingFile(path):
    """
    Reads an ANSYS .mapping file that maps equation to corresponding node.

    The .mapping file is structured:
    column 0: equation      column 1: node      column 2: dof
    Example:

    1   5   UX
    2   5   UY
    3   5   UZ

    This means equation 1 is related to node 5, dof UX
    Relation node -> dof is
    node 1, UX      => dof 1
    node 1, UY      => dof 2
    node 1, UZ      => dof 3 
    node 2, ROTX    => dof 4 ...

    node 2, UX      => dof 7

    :param path: File path of ".mapping" file
    :param dofDict: Dictionary relating internal dof for a node to it corresponding numbering. 
    Example: "UX" = 1, "UY" = 2, "UZ" = 3, "ROTX" = 4
    :return Dictionary
    dict = {
        "dof": "equation"
    }
    where "dof" is in ascending order from node numbers, i.e.: 
    dof = 1 => Node 1, degree of freedom nr 1 (UX)
    dof = 2 => Node 1, degree of freedom nr 2 (UY)
    dof = 7 => Node 2, degree of freedom nr 1 (UX)
    dof = 12 => Node 2, degree of freedom nr 6 (ROTZ)
    """
    dofDict = {
        "UX": 1,
        "UY": 2,
        "UZ": 3, 
        "ROTX": 4,
        "ROTY": 5, 
        "ROTZ": 6
    }
    dict = {}
    file = open(path, "r")
    line = file.readline()      # Skipping first line
    lines = file.readlines()
    for line in lines:
        line = line.split()
        eq = int(line[0])
        node = int(line[1])
        dofText = line[2]
        dof = node2dof(node, dofDict, 6, dofText)            # Currently hard-typing in 6 dofs
        dict[dof] = eq
    return dict

def node2dof(node, dofDict, dofs, dofText):
    """
    Relates node number to dof number. 

    :param node: int of node number
    :param dofs: Number of dofs for the node
    :param dofText: string of dof in ANSYS .mapping file. Ex.: "UX", "UY",
    """
    dof = (node-1)*dofs + dofDict[dofText]
    return dof





def nodes2Dofs(nodes, dofs):
    """
    # Relation dof -> node is
    # dof = (node - 1) * dofs + a
    # where a = [1,dofs]
    #
    # Example: 
    # nodes = [1, 5, 10] in elements with 6 dofs
    # => returnDofs = [1, 2, 3, 4, 5, 6, 25, 26, 27, 28, 29, 30, 55, 56, 57, 58, 59, 60]
    #
    # Example 2: 
    # nodes = [1, 3]
    # dofs = 2 
    # => returnDofs = [1, 2, 5, 6]
    """
    returnDofs = []
    for node in nodes:
        partDofs = range((node - 1) * dofs + 1, (node - 1) * dofs + (dofs + 1))
        returnDofs += partDofs
    return returnDofs


def dofs2Eqs(dofs, dict):
    """
    Returns a list with the corresponding equation for a node. 

    # Example:
    # dofs = [1, 2, 3]
    # dict = {
        1: 2,
        2: 3, 
        3: 1
        }
        returns [2, 3, 1]      

    :param dofs: List of dofs. 1-indexed. Length n
    :param dict: Dictionary coupling dof to equation like
        dict[dof] = eq
    :param return: List of equations for the specified dofs. Length n. 1-indexed.
    """
    eqs = []
    for dof in dofs:
        eq = dict[dof]
        eqs.append(eq)
    return eqs




def partitionMatrix(mat, eNodes, iNodes, dict, dofs, sorting="equation"):
    '''
    Creates a partitioned matrix on the form 
    M = 
    [[Mee, Mei], 
    [Mii, Mii]]
    
    :param mat: The Math.NET matrix to partition
    :param eNodes: list[int] of the selected external nodes. Indexing starting at 1. Sorted.    
    :param iNodes: list[int] of internal nodes. 1-indexed. Sorted.
    :param dict: Dictionary with the mapping of DOF to equation number in the matrix. 
        Example: dict[dof] = eq
        NOTE: dof and equation in the dictionary is 1-indexed!
    :param DOFs: Number of degrees of freedom per node. 
    :param sorting: How to sort the partitioned matrix. 
    sorting="equation" means the partitioned matrix is sorted after ascending equation numbers. This is what Fedem uses.
    This is the default selection.
    sorting="dof" means the partitioned matrix is sorted after ascending degree of freedom numbers. 

    :return partitioned matrix 
    '''
    
    # Getting the dof numbering for the external (e) and internal (i) nodes
    # Relation dof -> node is
    # dof = (node - 1) * dofs + a
    # where a = [1,dofs]
    # Example: 
    # eNodes = [1, 5, 10] in elements with 6 dofs
    # => eDofs = [1, 2, 3, 4, 5, 6, 31, 32, 33, 34, 35, 36, 61, 62, 63, 64, 65, 65]
    eDofs = nodes2Dofs(eNodes, dofs)
    iDofs = nodes2Dofs(iNodes, dofs)

    # Mapping to solver ordering: 
    eEqs = dofs2Eqs(eDofs, dict)            # equations corresponding to external nodes. 1-indexed.
    iEqs = dofs2Eqs(iDofs, dict)            # equations corresponding to internal nodes. 1-indexed.
    if (sorting=="equation"):       #eEqs and iEqs are already sorted by dof number.
        eEqs = sorted(eEqs)
        iEqs = sorted(iEqs)
    
    # Initializing partitioned matrix
    nRows = (len(eNodes) + len(iNodes)) * dofs    
    nCols = nRows
    matPart = la.Double.SparseMatrix(nRows, nCols)  # Initializing with zeroes

    # Moving rows
    row = 0
    for e in eEqs:
        matPart.SetRow(row,  mat.Row(e-1))
        row+=1
    for i in iEqs:
        matPart.SetRow(row,  mat.Row(i-1))
        row+=1
    matCopy = la.Double.SparseMatrix(nRows, nCols)       # Initializing with zeroes
    matPart.CopyTo(matCopy)     # Using a copy because values are accesed by reference.
    # Moving columns
    col = 0
    for e in eEqs:
        matPart.SetColumn(col,  matCopy.Column(e-1))
        col+=1
    for i in iEqs: 
        matPart.SetColumn(col,  matCopy.Column(i-1))
        col+=1
    return matPart





def readFedemMatrix(filename):
    """
    Reading a Fedem matrix to MathNETmatrix for easier comparison with ANSYS matrices.
    Written to import the full mass and stiffness matrix from Fedem

    :param filename: Path to matrix. 
    :return a sparse Math.NET matrix with contents of the matrix in path
    """
    file = open(filename, "r")      # Reading data
    line = file.readline()
    line = line.split()
    nRows = int(line[0])            # Number of columns and rows are stored in the first line after the comments
    nCols = int(line[1])
    mat = la.Double.SparseMatrix(nRows, nCols) # Initializing with zeros
    
    row = 0
    col = 0
    content = file.readlines()
    for line in range(0, len(content)):
        splitted = content[line].split()
        if (len(splitted) != 0):
            if (content[line].split()[0].startswith("+++")):
                row = int(content[line].split()[2]) - 1
                colLine = content[line + 1].split()
                valLine = content[line + 2].split()
                col = int(colLine[0]) - 1
                val = float(valLine[0])
                mat[row, col] = val
    file.close()
    return mat

def readFedemVector(path):
    """
    Reads vector from fedem_reducer.res file for comparison with vectors from ANSYS
    Reads gravity vectors, as well as BMAT.res and PHI.res

    :param path: Path to Fedem vector to read. 
    :return Sparse math.NET matrix of the Fedem vector in path
    """
    file = open(path, "r")
    lines = file.readlines()
    firstLine = lines[0].split()
    nRows = int(firstLine[0])
    nCols = int(firstLine[1])
    mat = la.Double.SparseMatrix(nRows, nCols)    # Initializing with zeros

    colIndexes = []
    nextNumCols = 0
    currNumCols = 0
    row = 0
    col = 0
    startCol = 0
    for i in range(0, len(lines)):
        col = startCol
        line = lines[i].split()
        if (i != (len(lines) - 1)):             # if not at end of file
            nextNumCols = len(lines[i +  1].split())
        currNumCols = len(line)
        if (currNumCols > nextNumCols):             # if new columns
            startCol = int(lines[i + 1].split()[0]) - 1
            continue
        row = int(line[0]) - 1
        vals = list(map(float, line))[1:]
        for val in vals:
            mat[row, col] = val
            col += 1
    file.close()
    return mat


def readMMFMatrix(filename, format):
    '''
    Reading a dense or sparse Matrix Market file format into a Math.NET array. 
    NOTE: 
    - The matrices are saved column first! 
    - Matrices created from the .SUB binary ANSYS file are saved with the full (n x n) matrix. 
        This is the case for M_red and K_red files
    - Matrices created from the .FULL binary ANSYS file have only the lower triangular part of the matrix saved.

    :param filename: String of file name to the .mmf file. Dense .mmf matrices are saved column-first!
    :param format: Format of the MMF matrix. "sub" for the reduced matrices. "full" for the full matrices. 
    :return  Dense Math.NET matrix read from filename. Saved row-first.
    '''
    isFirstLine = False
    nRows = 0
    nCols = 0
    file = open(filename, "r")      # Reading data
    #
    # Reading the lines defining dimensions of the matrix
    #
    dimLine = None
    while (isFirstLine == False):
        line = file.readline()
        if (not line.startswith('%')):
            isFirstLine = True
            line = line.split()
            nRows = int(line[0])            # Number of columns and rows are stored in the first line after the comments
            nCols = int(line[1])
            dimLine = line          # Saving for use in the next for loop

    # Initializing Math.NET matrix
    mat = la.Double.SparseMatrix(nRows, nCols)
    isDenseFull = False
    isSparseFull = False
    isSparseSub = False

    if (len(dimLine) == 2 and format=="full"):
        isDenseFull = True
    if (len(dimLine) == 3 and format=="full"):
        isSparseFull = True
    if (len(dimLine) == 3 and format=="sub"):
        isSparseSub = True


    # Filling Math.NET matrix with values from the file
    row = 0
    col = 0
    for line in file:
        if (isDenseFull):         # If we are reading a dense MMF matrix from ANSYS SUB or full file
            val = float(line)
            mat[row, col] = val

            if(row != nRows - 1 ):     # Updating counting variables
                row += 1
            else:
                col += 1
                row = 0
        if (isSparseSub or isSparseFull):       # We are reading a sparse MMF matrix from ANSYS FULL file
            line = line.split()
            i = int(line[0]) - 1            # Reading indexes. Is 1-indexed in the MMF file
            j = int(line[1]) - 1
            val = float(line[2])
            mat[i, j] = val             
            mat[j, i] = val             # sparse MMF matrix is saved in lower triangular form
    file.close()
    return mat




def getRBE2Nodes(load):
    '''
    Imports RBE2 nodes from NASTRAN file and returns list node ids of RBE2 nodes. 
    "CompID" is the string from the Workbench site, when pressing on external model, under "General" tab. Typically "Setup X"
    "Identifier" is the string under identifier tab when double-clicking the external model. Typically "File1".
    :param load: The current load object
    
    :return List[int] Sorted list with integer nodal ids of found RBE2 nodes in NASTRAN file.
    '''
    compID = load.Properties["Generation"].Properties["Geometry"].Properties["CompID"].Value
    identifier = load.Properties["Generation"].Properties["Geometry"].Properties["Identifier"].Value
    commands = ExtAPI.DataModel.Project.Model.GetFECommandsRepository(compID, identifier)
    rbe2Comms = commands.GetCommandsByName("RBE2")
    rbe2Count = rbe2Comms.Count
    nodeList = []
    for i in range (0, rbe2Count):
        cmd = rbe2Comms[i+1]
        nodeList.append(int(cmd.GetArgument[2]))        # Node ID is on argument index two. Convert to int.  
    return sorted(nodeList)
        
    
class Element():
    """
    Class used for storing info on elements in a mesh
    """

    def __init__(self, id, nodes):
        self._nodes = nodes        # list[Node] connected to the element
        self._id = id           # Element id
        

    def getNodes(self):
        return self._nodes

    def getId(self):
        return self._id




class Node:
    """
    Data storage class for one node from a mesh
    """

    def __init__(self, id, dofs, elements, x, y, z):
        """
        :param id: Int of nodal id
        :param dofs: Int of number of dofs for the node
        :param elements: List[int] of element ids to the elements connected to the node
        :param x: x coordinate
        :param y: y coordinate
        :param z: z coordinate
        """
        self._id = id
        self._dofs = dofs
        self._elements = elements
        self._x = x
        self._y = y
        self._z = z


    def getId():
        return self._id

    def getDofs():
        return self._dofs

    def getElements():
        return self._elements



class SolverData:
    #
    # Class used to store data from the solver and ModRed application.
    #

    def __init__(self, analysis, load):
        #
        # Collecting user input
        #
        self._dofs = 6         # Currently hard-typing in number of dofs
        self._filename = load.Properties["Method"].Properties["ReductionMethod"].Properties["Name"].Value  # string
        self._workingDir = analysis.WorkingDir
        self._fedemDir = load.Properties["Export"].Properties["IsExport"].Properties["FileDir"].Value
        self._testResourcesDir = ExtAPI.ExtensionManager.CurrentExtension.InstallDir + "\\test-resources"
        self._eNodes = getENodes(load)         # List[int] of selected external nodes. Sorted 
        self._cModes = int(load.Properties["Method"].Properties["ReductionMethod"].Properties["Nmode"].Value)     # Number of component modes
        
        self._elements = self.getElementsFromMesh(analysis.MeshData)           #List[Element]
        self._nodes = self.getNodesFromElements(self._elements)     #list[Node]
        
        self._nRows = len(self._nodes) * self._dofs      # Number of rows in the full mass matrix
        self._nCols = self._nRows         # Quadratic matrix



    def saveObject(self, obj, fname):
        """
        Saves the current object to _fedemDir
        """
        with open(self._testResourcesDir + "\\" + fname + ".p", "wb") as file: 
            pickle.dump(obj, file, pickle.HIGHEST_PROTOCOL)

    def getNodes(self):
        """
        Return sort list[Node] of nodes in the model. Sorted ascending by id
        """
        return sorted(self._nodes, key=lambda x: x._id)


    def getElements(self):
        """
        Return sorted list[Element] of elements in the model. Sorted ascending by id
        """
        return sorted(self._elements, key=lambda x: x._id)


    def getNodesFromElements(self, elements):
        """
        Creates list[Node] of nodes connected to the element.

        :param elements: List[Element]
        :return List[Node] with all nodes in the model
        """
        nodes = []
        for e in elements:
            for n in e.getNodes():
                if not any(x._id == n._id for x in nodes):
                    nodes.append(n)
        return nodes

    def getNodeIds(self):
        """
        :return Sorted list[int] of all node ids in the model
        """
        ids = []
        for node in self._nodes:
            ids.append(node._id)
        return sorted(ids)


    def getElementsFromMesh(self, mesh):
        """
        Returns list of custom Element objects containing all elements in the model.

        :param mesh: ANSYS MeshData object
        :return List[Element]   
        """
        returnElements = []
        ansysElements = mesh.Elements
        for e in ansysElements:
            id = e.Id

            nodeList = []
            for node in e.Nodes:
                nodeId = node.Id
                elements = node.ConnectedElementIds             #List[int] of ids of the connected elements
                node = Node(nodeId, 6, elements, node.X, node.Y, node.Z)            # Currently hard-typing in dofs per node. Should be changed to be read from ANSYS API
                nodeList.append(node)
            
            element = Element(id, nodeList)
            returnElements.append(element)
        return returnElements



class Results:
    """
    Class for calculating results from the model reduction extension
    """
    def __init__(self):
        self._resourcesFolder = os.path.dirname(os.path.realpath(__file__)) + "\\test-resources\\"



    def quadraticSumDifference(self, mat, refMat):
        """
        Quadratic sum of the difference of all elements in n x n matrix

        :param mat: Matrix that will be compared
        :param refMat: Matrix used as reference for comparison
        :return percentage difference on quadratic sum of mat, compared to refMat
        """
        matSum = self.matrixSum(mat)
        refMatSum = self.matrixSum(refMat)
        squared = (matSum - refMatSum) * (matSum - refMatSum)
        diff = math.sqrt(squared)
        percentage = diff/refMatSum       
        return percentage


    def quadraticSumDifferenceOfVector(self, vec, refVec):
        """
        Quadratic sum of the difference of two vectors of size (n x 1)

        :param : Vector that will be compared
        :param refVec: Vector used as reference for comparison
        :return percentage difference on quadratic sum of vec, compared to refVec
        """
        vecSum = vec.Sum()
        refVecSum = refVec.Sum()
        squared = (vecSum - refVecSum) * (vecSum - refVecSum)
        diff = math.sqrt(squared)
        percentage = diff/refVecSum       
        return percentage


    def matrixSum(self, mat):
        """
        Calculating the matrix sum of a mat.NET matrix
        :param mat: The matrix to calculate the sum of
        :return float Sum of all elements in the matrix
        """
        rowSums = mat.RowSums()
        return sum(rowSums)


    def calculateMassFromMassMatrix(self, mat, reduced=True, cmodes=2):
        """
        Calculates the mass of a mass matrix mat by applying unit translation: 
        m_tot = u^T * mat * u

        :param mat: The mass matrix
        :param reduced: Indicating if it is a reduced mass matrix. Defaults to True.
        :param cmodes: Number of component modes in the reduced mass matrix. Defaults to 2
        """
        if (reduced==True):
            nRows = mat.RowCount
            nNodes = int((nRows - cmodes) / 6)          # Currently assuming 6 DOFs per node
            u_x = ModRed.createUnitVector(1, nNodes, cmodes)
            u_y = ModRed.createUnitVector(2, nNodes, cmodes)
            u_z = ModRed.createUnitVector(3, nNodes, cmodes)
            mass_x = u_x.Transpose().Multiply(mat).Multiply(u_x)[0,0]
            mass_y = u_y.Transpose().Multiply(mat).Multiply(u_y)[0,0]
            mass_z = u_z.Transpose().Multiply(mat).Multiply(u_z)[0,0]
            return [mass_x, mass_y, mass_z]
        else: 
            nRows = mat.RowCount
            nNodes = int(nRows/6)       # currently assuming 6 dofs per node
            u_x = ModRed.createUnitVector(1, nNodes)
            u_y = ModRed.createUnitVector(2, nNodes)
            u_z = ModRed.createUnitVector(3, nNodes)
            mass_x = u_x.Transpose().Multiply(mat).Multiply(u_x)[0,0]
            mass_y = u_y.Transpose().Multiply(mat).Multiply(u_y)[0,0]
            mass_z = u_z.Transpose().Multiply(mat).Multiply(u_z)[0,0]
            return [mass_x, mass_y, mass_z]


